#include "Costmap.h"

#include <algorithm>
#include <cmath>
#include <limits>

#include <boost/geometry.hpp>
#include <boost/geometry/algorithms/detail/intersects/interface.hpp>

#include <Eigen/Core>
#include <Eigen/Geometry>

#include <IceUtil/Time.h>

#include <VirtualRobot/BoundingBox.h>
#include <VirtualRobot/CollisionDetection/CollisionChecker.h>
#include <VirtualRobot/CollisionDetection/CollisionModel.h>
#include <VirtualRobot/Visualization/CoinVisualization/CoinVisualizationFactory.h>
#include <VirtualRobot/XML/ObjectIO.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>
#include <ArmarXCore/core/logging/Logging.h>

#include "armarx/navigation/algorithms/types.h"
#include "armarx/navigation/algorithms/util.h"
#include <armarx/navigation/conversions/eigen.h>


namespace armarx::navigation::algorithms
{

    Costmap::Costmap(const Grid& grid,
                     const Parameters& parameters,
                     const SceneBounds& sceneBounds,
                     const std::optional<Mask>& mask) :
        grid(grid), mask(mask), sceneBounds(sceneBounds), parameters(parameters)
    {
        if (mask.has_value())
        {
            ARMARX_CHECK_EQUAL(grid.rows(), mask->rows());
            ARMARX_CHECK_EQUAL(grid.cols(), mask->cols());
        }
    }

    Costmap::Position
    Costmap::toPosition(const Costmap::Index& index) const
    {
        Eigen::Vector2f pos;
        pos.x() = sceneBounds.min.x() + index.x() * parameters.cellSize + parameters.cellSize / 2;
        pos.y() = sceneBounds.min.y() + index.y() * parameters.cellSize + parameters.cellSize / 2;

        return pos;
    }

    Costmap::Vertex
    Costmap::toVertex(const Eigen::Vector2f& v) const
    {
        const float vX =
            (v.x() - parameters.cellSize / 2 - sceneBounds.min.x()) / parameters.cellSize;
        const float vY =
            (v.y() - parameters.cellSize / 2 - sceneBounds.min.y()) / parameters.cellSize;

        const int iX = std::round(vX);
        const int iY = std::round(vY);

        const int iXSan = std::clamp<int>(iX, 0, grid.rows() - 1);
        const int iYSan = std::clamp<int>(iY, 0, grid.cols() - 1);

        // FIXME accept one cell off

        // ARMARX_CHECK_GREATER(iX, 0);
        // ARMARX_CHECK_GREATER(iY, 0);

        // ARMARX_CHECK_LESS_EQUAL(iX, grid.rows() - 1);
        // ARMARX_CHECK_LESS_EQUAL(iY, grid.cols() - 1);

        return Vertex{.index = Index{iXSan, iYSan}, .position = v};
    }

    const SceneBounds&
    Costmap::getSceneBounds() const noexcept
    {
        return sceneBounds;
    }

    Costmap::Index
    Costmap::minIndex() const noexcept
    {
        // index of the min element
        Grid::Index row = 0;
        Grid::Index col = 0;

        // value of the min element
        float minVal = std::numeric_limits<float>::max();

        for (int r = 0; r < grid.rows(); r++)
        {
            for (int c = 0; c < grid.cols(); c++)
            {
                if (mask.has_value())
                {
                    if (not mask.value()(r, c)) // skip invalid cells
                    {
                        continue;
                    }

                    const float currentVal = grid(r, c);
                    if (currentVal >= minVal)
                    {
                        minVal = currentVal;
                        row = r;
                        col = c;
                    }
                }
            }
        }


        return Index{row, col};
    }

    const Costmap::Parameters&
    Costmap::params() const noexcept
    {
        return parameters;
    }

    const Costmap::Grid&
    Costmap::getGrid() const
    {
        return grid;
    }

    bool
    Costmap::isInCollision(const Position& p) const
    {
        const auto v = toVertex(p);
        return grid(v.index.x(), v.index.y()) == 0.F;
    }

    bool
    Costmap::add(const Costmap& other)
    {
        const auto startIdx = toVertex(other.sceneBounds.min);

        // merge other grid into this one => use max costs

        // It might happen that this costmap is not large enough to cover the other costmap
        // In this case, only add the part to this costmap that is possible.
        const int rows = std::min(other.grid.rows(), grid.rows() - startIdx.index.x());
        const int cols = std::min(other.grid.cols(), grid.cols() - startIdx.index.y());

        ARMARX_INFO << "Adding other grid to region (" << startIdx.index.x() << ", "
                    << startIdx.index.y() << "), "
                    << " (" << startIdx.index.x() + rows << ", " << startIdx.index.y() + cols
                    << ")";

        // TODO if one of both matrices has mask, it has to be considered.
        ARMARX_TRACE;
        grid.block(startIdx.index.x(), startIdx.index.y(), rows, cols) =
            grid.block(startIdx.index.x(), startIdx.index.y(), rows, cols)
                .cwiseMax(other.grid.block(0, 0, rows, cols));

        if (not mask.has_value())
        {
            mask = Mask(grid.rows(), grid.cols());
            mask->setOnes();
        }

        Mask otherMask(grid.rows(), grid.cols());
        otherMask.setZero();

        if (other.mask.has_value())
        {
            otherMask.block(startIdx.index.x(), startIdx.index.y(), rows, cols) =
                otherMask.block(startIdx.index.x(), startIdx.index.y(), rows, cols)
                    .cwiseMin(other.mask->value());
        }
        else
        {
            otherMask.block(startIdx.index.x(), startIdx.index.y(), rows, cols).setOnes();
        }

        mask.value() = mask->cwiseMin(otherMask);

        // apply mask to grid => all invalid points will be 0
        ARMARX_TRACE;
        grid = grid.cwiseProduct(mask->cast<float>());

        return true;
    }


} // namespace armarx::navigation::algorithms
