/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "util.h"

#include <algorithm>
#include <iterator>

#include <opencv2/core/eigen.hpp>
#include <opencv2/imgproc.hpp>

#include <SimoxUtility/algorithm/apply.hpp>
#include <VirtualRobot/BoundingBox.h>
#include <VirtualRobot/CollisionDetection/CollisionModel.h>
#include <VirtualRobot/SceneObjectSet.h>
#include <VirtualRobot/Workspace/WorkspaceGrid.h>

#include "ArmarXCore/core/exceptions/local/ExpressionException.h"
#include "ArmarXCore/core/logging/Logging.h"

#include "armarx/navigation/algorithms/Costmap.h"
#include "armarx/navigation/algorithms/CostmapBuilder.h"
#include "armarx/navigation/algorithms/types.h"

namespace armarx::navigation::algorithms
{


    SceneBounds
    computeSceneBounds(const VirtualRobot::SceneObjectSetPtr& obstacles)
    {
        SceneBounds bounds;

        ARMARX_CHECK_NOT_NULL(obstacles);
        ARMARX_CHECK(not obstacles->getCollisionModels().empty());

        for (size_t i = 0; i < obstacles->getCollisionModels().size(); i++)
        {
            VirtualRobot::BoundingBox bb = obstacles->getCollisionModels()[i]->getBoundingBox();
            bounds.min.x() = std::min(bb.getMin().x(), bounds.min.x());
            bounds.min.y() = std::min(bb.getMin().y(), bounds.min.y());
            bounds.max.x() = std::max(bb.getMax().x(), bounds.max.x());
            bounds.max.y() = std::max(bb.getMax().y(), bounds.max.y());
        }

        return bounds;
    }


    SceneBounds
    toSceneBounds(const VirtualRobot::WorkspaceGrid::Extends& extends)
    {
        return {.min = Eigen::Vector2f{extends.minX, extends.minY},
                .max = Eigen::Vector2f{extends.maxX, extends.maxY}};
    }

    SceneBounds
    merge(const std::vector<SceneBounds>& sceneBounds)
    {
        SceneBounds bounds;

        const auto expandBounds = [&bounds](const SceneBounds& sb)
        {
            bounds.min.x() = std::min(bounds.min.x(), sb.min.x());
            bounds.min.y() = std::min(bounds.min.y(), sb.min.y());

            bounds.max.x() = std::max(bounds.max.x(), sb.max.x());
            bounds.max.y() = std::max(bounds.max.y(), sb.max.y());

            return bounds;
        };

        std::for_each(sceneBounds.begin(), sceneBounds.end(), expandBounds);

        return bounds;
    }


    Costmap
    toCostmap(const VirtualRobot::WorkspaceGrid& workspaceGrid)
    {
        const Costmap::Grid grid(workspaceGrid.getCells().x, workspaceGrid.getCells().y);
        const SceneBounds sceneBounds = toSceneBounds(workspaceGrid.getExtends());

        const Costmap::Parameters parameters{.binaryGrid = false,
                                             .cellSize = workspaceGrid.getDiscretizeSize()};

        return {grid, parameters, sceneBounds};
    }


    Costmap
    mergeUnaligned(const std::vector<Costmap>& costmaps,
                   const std::vector<float>& weights,
                   float cellSize)
    {
        ARMARX_CHECK_EQUAL(costmaps.size(), weights.size());

        // if unset, use the smallest cell size
        if (cellSize < 0)
        {
            const auto compCellSize = [](const Costmap& a, const Costmap& b) -> bool
            { return a.params().cellSize < b.params().cellSize; };

            cellSize =
                std::min_element(costmaps.begin(), costmaps.end(), compCellSize)->params().cellSize;
        }

        // scale each costmap
        std::vector<Costmap> scaledCostmaps = simox::alg::apply(
            costmaps,
            [&cellSize](const Costmap& costmap) -> Costmap { return scaleCostmap(costmap, cellSize); });

        // merge costmaps into one

        // - combine scene bounds
        std::vector<SceneBounds> sceneBoundsAll = simox::alg::apply(
            scaledCostmaps, [](const Costmap& costmap) { return costmap.getSceneBounds(); });

        const SceneBounds sceneBounds = merge(sceneBoundsAll);

        // - create grid
        const auto largeGrid = CostmapBuilder::createUniformGrid(
            sceneBounds, Costmap::Parameters{.binaryGrid = false, .cellSize = cellSize});

        Costmap largeCostmap(
            largeGrid, Costmap::Parameters{.binaryGrid = false, .cellSize = cellSize}, sceneBounds);

        // - add all costmaps to this one
        std::for_each(scaledCostmaps.begin(),
                      scaledCostmaps.end(),
                      [&largeCostmap](const Costmap& costmap) { largeCostmap.add(costmap); });

        // return merged grid
        return largeCostmap;
    }

    Costmap
    scaleCostmap(const Costmap& costmap, float cellSize)
    {
        const float scale = costmap.params().cellSize / cellSize;
        ARMARX_INFO << "Scaling grid by a factor of " << scale;

        cv::Mat src;
        cv::eigen2cv(costmap.getGrid(), src);

        cv::Mat dst;
        cv::resize(src, dst, cv::Size{0, 0}, scale, scale);

        // TODO properly use costmap.mask
        // cv::Mat mask = src.clone();
        // mask = cv::threshold(mask, mask, 0.5, 1.0, cv::THRESH_BINARY);
        // cv::resize(mask, mask, cv::Size{0, 0}, scale, scale);
        // dst = dst * mask;

        // TODO do the same with a mask

        Eigen::MatrixXf scaledGrid;
        cv::cv2eigen(dst, scaledGrid);

        ARMARX_INFO << "Original size (" << costmap.getGrid().rows() << ", " << costmap.getGrid().cols() << ")";
        ARMARX_INFO << "Resized to (" << scaledGrid.rows() << ", " << scaledGrid.cols() << ")";

        return {scaledGrid,
                Costmap::Parameters{.binaryGrid = false, .cellSize = cellSize},
                costmap.getSceneBounds()};
    }


} // namespace armarx::navigation::algorithms
