/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    Navigation::ArmarXObjects::NavigationMemory
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "NavigationMemory.h"

#include <filesystem>

#include <ArmarXCore/core/system/ArmarXDataPath.h>
#include <ArmarXCore/core/time/CycleUtil.h>
#include <ArmarXCore/libraries/DecoupledSingleComponent/Decoupled.h>

#include <RobotAPI/libraries/armem/client/query.h>
#include <RobotAPI/libraries/armem/core/operations.h>
#include <RobotAPI/libraries/armem/server/ltm/disk/Memory.h>
#include <RobotAPI/libraries/armem/server/wm/memory_definitions.h>

#include "Visu.h"
#include <armarx/navigation/core/Graph.h>
#include <armarx/navigation/core/aron/Graph.aron.generated.h>
#include <armarx/navigation/core/aron/Location.aron.generated.h>
#include <armarx/navigation/core/aron/Trajectory.aron.generated.h>
#include <armarx/navigation/core/aron/Twist.aron.generated.h>
#include <armarx/navigation/graph/constants.h>
#include <armarx/navigation/location/constants.h>


namespace armarx::navigation
{

    ARMARX_DECOUPLED_REGISTER_COMPONENT(NavigationMemory);


    armarx::PropertyDefinitionsPtr
    NavigationMemory::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr def =
            new ComponentPropertyDefinitions(getConfigIdentifier());

        // Publish to a topic (passing the TopicListenerPrx).
        // def->topic(myTopicListener);

        // Subscribe to a topic (passing the topic name).
        // def->topic<PlatformUnitListener>("MyTopic");

        // Use (and depend on) another component (passing the ComponentInterfacePrx).
        // def->component(myComponentProxy)


        workingMemory().name() = "Navigation";
        def->optional(properties.snapshotToLoad,
                      "p.snapshotToLoad",
                      "Memory snapshot to load at start up \n"
                      "(e.g. 'PriorKnowledgeData/navigation-graphs/snapshot').");

        def->optional(properties.locationGraph.visuLocations,
                      "p.locationGraph.visuLocation",
                      "Enable visualization of locations.");
        def->optional(properties.locationGraph.visuGraphEdges,
                      "p.locationGraph.visuGraphEdges",
                      "Enable visualization of navigation graph edges.");
        def->optional(properties.locationGraph.visuFrequency,
                      "p.locationGraph.visuFrequency",
                      "Visualization frequeny of locations and graph edges [Hz].");

        return def;
    }


    std::optional<std::filesystem::path>
    findSnapshotDirectory(const std::filesystem::path& specifiedPath, const std::string& memoryName)
    {
        std::vector<std::filesystem::path> candidates{
            specifiedPath,
            specifiedPath / memoryName,
        };
        for (const auto& candidate : candidates)
        {
            if (candidate.filename() == memoryName and std::filesystem::is_directory(candidate))
            {
                return candidate;
            }
        }
        return std::nullopt;
    }


    void
    NavigationMemory::onInitComponent()
    {
        // Topics and properties defined above are automagically registered.

        // Keep debug observer data until calling `sendDebugObserverBatch()`.
        // (Requies the armarx::DebugObserverComponentPluginUser.)
        // setDebugObserverBatchModeEnabled(true);

        workingMemory().addCoreSegment("Parameterization");

        workingMemory().addCoreSegment("Results_GlobalPlanner",
                                       navigation::core::arondto::Trajectory::toAronType());
        workingMemory().addCoreSegment("Results_LocalPlanner",
                                       navigation::core::arondto::Trajectory::toAronType());
        workingMemory().addCoreSegment("Results_TrajectoryController",
                                       navigation::core::arondto::Twist::toAronType());
        workingMemory().addCoreSegment("Results_SafetyController",
                                       navigation::core::arondto::Twist::toAronType());

        workingMemory().addCoreSegment("Events"); //, armem::example::ExampleData::toAronType());
        // workingMemory.addCoreSegment("Exceptions");  //, armem::example::ExampleData::toAronType());


        workingMemory().addCoreSegment(navigation::location::coreSegmentID.coreSegmentName,
                                       navigation::location::arondto::Location::toAronType());
        workingMemory().addCoreSegment(navigation::graph::coreSegmentID.coreSegmentName,
                                       navigation::core::arondto::Graph::toAronType());


        if (not properties.snapshotToLoad.empty())
        {
            std::string resolved = ArmarXDataPath::resolvePath(properties.snapshotToLoad);
            if (auto path = findSnapshotDirectory(resolved, workingMemory().name()))
            {
                // armem::wm::Memory memory = armem::server::ltm::disk::load(path.value());
                armem::server::ltm::disk::Memory ltm(path.value());
                armem::wm::Memory memory;
                ltm.loadAll(memory);

                ARMARX_INFO << "Loaded " << armem::print(memory);
                workingMemory().update(armem::toCommit(memory));
            }
        }
    }


    void
    NavigationMemory::onConnectComponent()
    {
        // Do things after connecting to topics and components.


        /* (Requies the armarx::DebugObserverComponentPluginUser.)
        // Use the debug observer to log data over time.
        // The data can be viewed in the ObserverView and the LivePlotter.
        // (Before starting any threads, we don't need to lock mutexes.)
        {
            setDebugObserverDatafield("numBoxes", properties.numBoxes);
            setDebugObserverDatafield("boxLayerName", properties.boxLayerName);
            sendDebugObserverBatch();
        }
        */

        // Setup the remote GUI.
        {
            createRemoteGuiTab();
            RemoteGui_startRunningTask();
        }

        tasks.visuTask = new SimpleRunningTask<>([this]() { this->visuRun(); }, "Visualization");
        tasks.visuTask->start();
    }


    void
    NavigationMemory::onDisconnectComponent()
    {
    }


    void
    NavigationMemory::onExitComponent()
    {
    }


    std::string
    NavigationMemory::getDefaultName() const
    {
        return "NavigationMemory";
    }


    void
    NavigationMemory::createRemoteGuiTab()
    {
        using namespace armarx::RemoteGui::Client;

        // Setup the widgets.
        tab.locationGraph.setup(*this);


        // Setup the layout.
        VBoxLayout root = {tab.locationGraph.group, VSpacer()};
        RemoteGui_createTab(getName(), root, &tab);
    }


    void
    NavigationMemory::RemoteGui_update()
    {
        tab.locationGraph.update(*this);
    }


    void
    NavigationMemory::RemoteGuiTab::LocationGraph::setup(NavigationMemory& owner)
    {
        using namespace armarx::RemoteGui::Client;
        GridLayout grid;
        int row = 0;
        {
            visuLocations.setValue(owner.properties.locationGraph.visuLocations);
            visuGraphEdges.setValue(owner.properties.locationGraph.visuLocations);

            grid.add(Label("Visualize Locations"), {row, 0}).add(visuLocations, {row, 1});
            ++row;

            grid.add(Label("Visualize Graph Edges"), {row, 0}).add(visuGraphEdges, {row, 1});
            ++row;
        }

        group = GroupBox({grid});
    }


    void
    NavigationMemory::RemoteGuiTab::LocationGraph::update(NavigationMemory& owner)
    {
        if (visuLocations.hasValueChanged() or visuGraphEdges.hasValueChanged())
        {
            std::scoped_lock lock(owner.propertiesMutex);
            owner.properties.locationGraph.visuLocations = visuLocations.getValue();
            owner.properties.locationGraph.visuGraphEdges = visuGraphEdges.getValue();
        }
    }


    void
    NavigationMemory::visuRun()
    {
        memory::Visu visu{arviz,
                          workingMemory().getCoreSegment(navigation::location::coreSegmentID),
                          workingMemory().getCoreSegment(navigation::graph::coreSegmentID)};

        Properties::LocationGraph p;
        {
            std::scoped_lock lock(propertiesMutex);
            p = properties.locationGraph;
        }

        CycleUtil cycle(static_cast<int>(1000 / p.visuFrequency));
        while (tasks.visuTask and not tasks.visuTask->isStopped())
        {
            {
                std::scoped_lock lock(propertiesMutex);
                p = properties.locationGraph;
            }

            std::vector<viz::Layer> layers;

            // Locations
            visu.drawLocations(layers, p.visuLocations);

            // Graph Edges
            visu.drawGraphs(layers, p.visuGraphEdges);

            arviz.commit(layers);

            cycle.waitForCycleDuration();
        }
    }


} // namespace armarx::navigation
